/// <reference types="webxr" />
import { Component, ComponentProto, ComponentConstructor, CustomParameter, Texture, Object as Object3D, Scene, Physics } from './wonderland.js';
import { WASM } from './wasm.js';
/**
 * Callback triggered when the session starts.
 *
 * Used in {@link WonderlandEngine.onXRSessionStart}.
 *
 * @param session The session that started
 */
export type XrSessionStartCallback = (session: XRSession, mode: XRSessionMode) => void;
/**
 * Callback triggered for supported / unsupported session.
 *
 * Used in {@link WonderlandEngine.onXRSupported}.
 *
 * @param type Type of session which is supported/not supported. Either `'vr'` or `'ar'`.
 * @param supported Whether given session type is supported.
 */
export type XrSupportCallback = (type: string, supported: boolean) => void;
/**
 * Callback triggered when the scene is loaded.
 *
 * Used in {@link WonderlandEngine.onSceneLoaded}.
 */
export type SceneLoadedCallback = () => void;
export type TextureCache = {
    [key: string]: Texture | ((filename: string, crossOrigin?: string) => Promise<Texture>);
    /**
     * Load an image from URL as {@link Texture}.
     *
     * @param filename URL to load from.
     * @param crossOrigin Cross origin flag for the image object.
     * @returns Loaded texture.
     */
    load: (filename: string, crossOrigin?: string) => Promise<Texture>;
};
/**
 * Main Wonderland Engine instance.
 *
 * Controls the canvas, rendering, and JS <-> WASM communication.
 */
export declare class WonderlandEngine {
    #private;
    /** List of functions to call if a WebXR session is started. */
    readonly onXRSessionStart: XrSessionStartCallback[];
    /** List of functions to call if a WebXR session ends. */
    readonly onXRSessionEnd: (() => void)[];
    /**
     * Whether AR is supported by the browser.
     *
     * `undefined` until support could be determined.
     */
    readonly arSupported: boolean | undefined;
    /**
     * Whether VR is supported by the browser.
     *
     * `undefined` until support could be determined.
     */
    readonly vrSupported: boolean | undefined;
    /**
     * List of functions to call once VR/AR support has been determined.
     *
     * Will be called once for AR and once for VR independent of support for each.
     * This allows you to notify the user of both cases: support and missing support of XR.
     * See the `supported` parameter of the callback, which indicates support.
     */
    readonly onXRSupported: XrSupportCallback[];
    /**
     * List of functions to call once the main scene has been loaded.
     */
    readonly onSceneLoaded: SceneLoadedCallback[];
    /**
     * Current main scene.
     */
    readonly scene: Scene;
    /**
     * Canvas element that Wonderland Engine renders to.
     */
    canvas: HTMLCanvasElement | null;
    /**
     * Access to the textures managed by Wonderland Engine.
     */
    textures: TextureCache;
    private _componentCache;
    private readonly _objectCache;
    /**
     * Create a new engine instance.
     *
     * @param wasm Wasm bridge instance
     *
     * @hidden
     */
    constructor(wasm: WASM);
    /**
     * Start the engine if it's not already running.
     *
     * When using the {@link loadRuntime} function, this method is called
     * automatically.
     */
    start(): void;
    /**
     * Register a custom JavaScript component type.
     *
     * ```js
     * registerComponent('my-new-type', {
     *    myParam: {type: Type.Float, default: 42.0},
     * }, {
     *    init: function() {},
     *    start: function() {},
     *    update: function(dt) {},
     *    onActivate: function() {},
     *    onDeactivate: function() {},
     *    onDestroy: function() {},
     * });
     * ```
     *
     * @param name Name of the component.
     * @param params Dict of param names to {@link CustomParameter}.
     * @param object Object containing functions for the component type.
     * @deprecated Use {@link WonderlandEngine.registerComponent:CLASSES} instead.
     */
    registerComponent(name: string, params: {
        [key: string]: CustomParameter;
    }, object: ComponentProto): void;
    /**
     * Register a custom JavaScript component type.
     *
     * You can register a component directly using a class inheriting from {@link Component}:
     *
     * ```js
     * import { Component, Type } from '@wonderlandengine/api';
     *
     * export class MyComponent extends Component {
     *     static TypeName = 'my-component';
     *     static Properties = {
     *         myParam: {type: Type.Float, default: 42.0},
     *     };
     *     init() {}
     *     start() {}
     *     update(dt) {}
     *     onActivate() {}
     *     onDeactivate() {}
     *     onDestroy() {}
     * });
     *
     * // Here, we assume we have an engine already instantiated.
     * // In general, the registration occurs in the `index.js` file in your
     * // final application.
     * engine.registerComponent(MyComponent);
     * ```
     *
     * {@label CLASSES}
     * @param classes Custom component(s) extending {@link Component}.
     *
     * @since 1.0.0
     */
    registerComponent(...classes: ComponentConstructor[]): void;
    /**
     * Request a XR session.
     *
     * @note Please use this call instead of directly calling `navigator.xr.requestSession()`.
     * Wonderland Engine requires to be aware that a session is started, and this
     * is done through this call.
     *
     * @param mode The XR mode.
     * @param features An array of required features, e.g., `['local-floor', 'hit-test']`.
     * @param optionalFeatures An array of optional features, e.g., `['bounded-floor', 'depth-sensing']`.
     * @returns A promise resolving with the `XRSession`, a string error message otherwise.
     */
    requestXRSession(mode: XRSessionMode, features: string[], optionalFeatures?: string[]): Promise<XRSession>;
    /**
     * Wrap an object ID using {@link Object}.
     *
     * @note This method performs caching and will return the same
     * instance on subsequent calls.
     *
     * @param objectId ID of the object to create.
     *
     * @returns The object
     */
    wrapObject(objectId: number): Object3D;
    /**
     * WebAssembly bridge.
     *
     * @note Use with care. This object is used to communicate
     * with the WebAssembly code throughout the api.
     *
     * @hidden
     */
    get wasm(): WASM;
    /** Current WebXR session or `null` if no session active. */
    get xrSession(): XRSession | null;
    /** Current WebXR frame or `null` if no session active. */
    get xrFrame(): XRFrame | null;
    /** Current WebXR frame or `null` if no session active. */
    get xrBaseLayer(): XRProjectionLayer | null;
    /** Current WebXR WebGLFramebuffr or `null` if no session active. */
    get xrFramebuffer(): WebGLFramebuffer | null;
    /** Physics manager, only available when physx is enabled in the runtime. */
    get physics(): Physics | null;
    /**
     * Initialize the engine.
     *
     * @note Should be called after the WebAssembly is fully loaded.
     *
     * @hidden
     */
    _init(): void;
    /**
     * Reset the runtime state, including:
     *     - Component cache
     *     - Images
     *     - Callbacks
     *
     * @note This api is meant to be used internally.
     *
     * @hidden
     */
    _reset(): void;
    /**
     * Retrieves a component instance if it exists, or create and cache
     * a new one.
     *
     * @note This api is meant to be used internally. Please have a look at
     * {@link Object3D.addComponent} instead.
     *
     * @param type component type name
     * @param componentType Component manager index
     * @param componentId Component id in the manager
     *
     * @returns JavaScript instance wrapping the native component
     *
     * @hidden
     */
    _wrapComponent(type: string, componentType: number, componentId: number): Component | null;
}
