import { CollisionComponent, AnimationComponent, LightComponent, MeshComponent, PhysXComponent, InputComponent, ViewComponent, TextComponent, Texture, Object as Object3D, Scene, Type, Physics, } from './wonderland.js';
import { isString } from './utils/object.js';
/**
 * Default component parameter value per type.
 */
const _componentDefaults = new Map([
    [Type.Bool, false],
    [Type.Int, 0],
    [Type.Float, 0.0],
    [Type.String, ''],
    [Type.Enum, undefined],
    [Type.Object, null],
    [Type.Mesh, null],
    [Type.Texture, null],
    [Type.Material, null],
    [Type.Animation, null],
    [Type.Skin, null],
    [Type.Color, [0.0, 0.0, 0.0, 1.0]],
]);
/**
 * Setup the defaults value of the properties on a given
 * component class.
 *
 * @param ctor The component class
 */
function _setupDefaults(ctor) {
    for (const name in ctor.Properties) {
        const p = ctor.Properties[name];
        /* Convert enum default string to an index. We need to check if it's
         * already been converted so we don't try to look up the index. */
        if (p.type === Type.Enum && typeof p.default !== 'number') {
            /* Matches the editor behavior for packaged components. Default
             * value is a string, but is converted to an index. If there is no
             * default string, it becomes index 0. */
            if (p.values?.length) {
                p.default = Math.max(p.values.indexOf(p.default), 0);
            }
            else {
                p.default = undefined;
            }
        }
        else {
            p.default = p.default ?? _componentDefaults.get(p.type);
        }
        ctor.prototype[name] = p.default;
    }
}
let vrSupported = null;
let arSupported = null;
const onVRSupportEvent = [];
const onARSupportEvent = [];
/**
 * Check whether XR is supported and store the result in the global space.
 *
 * @todo Remove at 1.0.0. This can be moved in `loadRuntime` using a `Promise.all`.
 * Then, the `onXRSupported` callback becomes deprecated.
 */
function checkWebXRSupport() {
    if (typeof navigator === 'undefined' || !navigator.xr) {
        vrSupported = false;
        arSupported = false;
        return;
    }
    navigator.xr.isSessionSupported('immersive-vr').then(function (supported) {
        vrSupported = supported;
        for (const cb of onVRSupportEvent)
            cb(supported);
    });
    navigator.xr.isSessionSupported('immersive-ar').then(function (supported) {
        arSupported = supported;
        for (const cb of onARSupportEvent)
            cb(supported);
    });
}
/**
 * Main Wonderland Engine instance.
 *
 * Controls the canvas, rendering, and JS <-> WASM communication.
 */
export class WonderlandEngine {
    /** List of functions to call if a WebXR session is started. */
    onXRSessionStart = [];
    /** List of functions to call if a WebXR session ends. */
    onXRSessionEnd = [];
    /**
     * Whether AR is supported by the browser.
     *
     * `undefined` until support could be determined.
     */
    arSupported = undefined;
    /**
     * Whether VR is supported by the browser.
     *
     * `undefined` until support could be determined.
     */
    vrSupported = undefined;
    /**
     * List of functions to call once VR/AR support has been determined.
     *
     * Will be called once for AR and once for VR independent of support for each.
     * This allows you to notify the user of both cases: support and missing support of XR.
     * See the `supported` parameter of the callback, which indicates support.
     */
    onXRSupported = [];
    /**
     * List of functions to call once the main scene has been loaded.
     */
    onSceneLoaded = [];
    /**
     * Current main scene.
     */
    scene = null;
    /**
     * Canvas element that Wonderland Engine renders to.
     */
    canvas = null;
    /**
     * Access to the textures managed by Wonderland Engine.
     */
    textures;
    /* Component class instances per type to avoid GC */
    _componentCache = {};
    /* Object class instances per type to avoid GC */
    _objectCache = [];
    /**
     * WebAssembly bridge.
     *
     * @hidden
     */
    #wasm;
    /**
     * Physics manager, only available when physx is enabled in the runtime.
     *
     * @hidden
     */
    #physics = null;
    /**
     * Create a new engine instance.
     *
     * @param wasm Wasm bridge instance
     *
     * @hidden
     */
    constructor(wasm) {
        this.#wasm = wasm;
        this.#wasm['_setEngine'](this); /* String lookup to bypass private. */
        this.textures = {
            /* Backward compatibility. @todo: Remove at 1.0.0. */
            load: (filename, crossOrigin) => {
                let image = new Image();
                if (crossOrigin !== undefined) {
                    image.crossOrigin = crossOrigin;
                }
                image.src = filename;
                return new Promise((resolve, reject) => {
                    image.onload = () => {
                        let texture = new Texture(this, image);
                        if (!texture.valid) {
                            reject('Failed to add image ' +
                                image.src +
                                ' to texture atlas. Probably incompatible format.');
                        }
                        resolve(texture);
                    };
                    image.onerror = function () {
                        reject('Failed to load image. Not found or no read access');
                    };
                });
            },
        };
        this._componentCache = {};
        this._objectCache.length = 0;
        const onArSupported = (supported) => {
            const index = onARSupportEvent.indexOf(onArSupported);
            if (index >= 0)
                onARSupportEvent.splice(index, 1);
            this.arSupported = supported;
            for (const cb of this.onXRSupported) {
                cb('ar', supported);
            }
        };
        onARSupportEvent.push(onArSupported);
        const onVrSupported = (supported) => {
            const index = onVRSupportEvent.indexOf(onVrSupported);
            if (index >= 0)
                onVRSupportEvent.splice(index, 1);
            this.vrSupported = supported;
            for (const cb of this.onXRSupported) {
                cb('vr', supported);
            }
        };
        onVRSupportEvent.push(onVrSupported);
        /* If the global scope already knows about AR / VR support,
         * directly trigger the callbacks. */
        if (arSupported !== null)
            onArSupported(arSupported);
        if (vrSupported !== null)
            onVrSupported(vrSupported);
    }
    /**
     * Start the engine if it's not already running.
     *
     * When using the {@link loadRuntime} function, this method is called
     * automatically.
     */
    start() {
        this.wasm._wl_application_start();
    }
    /** @overload */
    registerComponent(...args) {
        if (isString(args[0])) {
            /* Registration is using `name`, `params`, and `object`. */
            const typeIndex = this.wasm._registerComponentLegacy(args[0], args[1], args[2]);
            const ctor = this.wasm._componentTypes[typeIndex];
            _setupDefaults(ctor);
            return;
        }
        for (const arg of args) {
            const typeIndex = this.wasm._registerComponent(arg);
            const ctor = this.wasm._componentTypes[typeIndex];
            _setupDefaults(ctor);
        }
    }
    /**
     * Request a XR session.
     *
     * @note Please use this call instead of directly calling `navigator.xr.requestSession()`.
     * Wonderland Engine requires to be aware that a session is started, and this
     * is done through this call.
     *
     * @param mode The XR mode.
     * @param features An array of required features, e.g., `['local-floor', 'hit-test']`.
     * @param optionalFeatures An array of optional features, e.g., `['bounded-floor', 'depth-sensing']`.
     * @returns A promise resolving with the `XRSession`, a string error message otherwise.
     */
    requestXRSession(mode, features, optionalFeatures) {
        return this.#wasm.webxr_request_session_func(mode, features, optionalFeatures ?? []);
    }
    /**
     * Wrap an object ID using {@link Object}.
     *
     * @note This method performs caching and will return the same
     * instance on subsequent calls.
     *
     * @param objectId ID of the object to create.
     *
     * @returns The object
     */
    wrapObject(objectId) {
        const cache = this._objectCache;
        const o = cache[objectId] || (cache[objectId] = new Object3D(this, objectId));
        o.objectId = objectId;
        return o;
    }
    /* Public Getters & Setter */
    /**
     * WebAssembly bridge.
     *
     * @note Use with care. This object is used to communicate
     * with the WebAssembly code throughout the api.
     *
     * @hidden
     */
    get wasm() {
        return this.#wasm;
    }
    /** Current WebXR session or `null` if no session active. */
    get xrSession() {
        return this.#wasm.webxr_session;
    }
    /** Current WebXR frame or `null` if no session active. */
    get xrFrame() {
        return this.#wasm.webxr_frame;
    }
    /** Current WebXR frame or `null` if no session active. */
    get xrBaseLayer() {
        return this.#wasm.webxr_baseLayer;
    }
    /** Current WebXR WebGLFramebuffr or `null` if no session active. */
    get xrFramebuffer() {
        if (!Array.isArray(this.wasm.webxr_fbo)) {
            return this.wasm.GL.framebuffers[this.wasm.webxr_fbo];
        }
        /* For now, we only use a single framebuffer. */
        return this.wasm.GL.framebuffers[this.wasm.webxr_fbo[0]];
    }
    /** Physics manager, only available when physx is enabled in the runtime. */
    get physics() {
        return this.#physics;
    }
    /* Internal-Only Methods */
    /**
     * Initialize the engine.
     *
     * @note Should be called after the WebAssembly is fully loaded.
     *
     * @hidden
     */
    _init() {
        this.scene = new Scene(this);
        /* For internal testing, we provide compatibility with DOM-less execution */
        this.canvas =
            typeof document === 'undefined'
                ? null
                : document.getElementById('canvas');
        /* Setup the error handler. This is used to to manage native errors. */
        this.#wasm._wl_set_error_callback(this.#wasm.addFunction((messagePtr) => {
            throw new Error(this.#wasm.UTF8ToString(messagePtr));
        }, 'vi'));
        this.#physics = null;
        if (this.#wasm.withPhysX) {
            /* Setup the physics callback. */
            const physics = new Physics(this);
            this.#wasm._wl_physx_set_collision_callback(this.#wasm.addFunction((a, index, type, b) => {
                const callback = physics._callbacks[a][index];
                const component = new PhysXComponent(this, this.wasm._typeIndexFor('physx'), b);
                callback(type, component);
            }, 'viiii'));
            this.#physics = physics;
        }
        /*
         * This is done here for backward compatibility.
         *
         * @todo 1.0.0 Move this to the top-level.
         */
        checkWebXRSupport();
    }
    /**
     * Reset the runtime state, including:
     *     - Component cache
     *     - Images
     *     - Callbacks
     *
     * @note This api is meant to be used internally.
     *
     * @hidden
     */
    _reset() {
        this._componentCache = {};
        this._objectCache.length = 0;
        this.scene.reset();
        this.wasm.reset();
    }
    /**
     * Retrieves a component instance if it exists, or create and cache
     * a new one.
     *
     * @note This api is meant to be used internally. Please have a look at
     * {@link Object3D.addComponent} instead.
     *
     * @param type component type name
     * @param componentType Component manager index
     * @param componentId Component id in the manager
     *
     * @returns JavaScript instance wrapping the native component
     *
     * @hidden
     */
    _wrapComponent(type, componentType, componentId) {
        if (componentId < 0)
            return null;
        /* @todo: extremely slow in JS to do that... Better to use a Map or allocate the array. */
        const c = this._componentCache[componentType] ||
            (this._componentCache[componentType] = []);
        if (c[componentId]) {
            return c[componentId];
        }
        let component;
        if (type == 'collision') {
            component = new CollisionComponent(this, componentType, componentId);
        }
        else if (type == 'text') {
            component = new TextComponent(this, componentType, componentId);
        }
        else if (type == 'view') {
            component = new ViewComponent(this, componentType, componentId);
        }
        else if (type == 'mesh') {
            component = new MeshComponent(this, componentType, componentId);
        }
        else if (type == 'input') {
            component = new InputComponent(this, componentType, componentId);
        }
        else if (type == 'light') {
            component = new LightComponent(this, componentType, componentId);
        }
        else if (type == 'animation') {
            component = new AnimationComponent(this, componentType, componentId);
        }
        else if (type == 'physx') {
            component = new PhysXComponent(this, componentType, componentId);
        }
        else {
            const typeIndex = this.wasm._componentTypeIndices[type];
            const constructor = this.wasm._componentTypes[typeIndex];
            component = new constructor(this);
        }
        /* Sets the manager and identifier from the outside, to
         * simplify the user's constructor. */
        /* @ts-ignore */
        component._engine = this;
        component._manager = componentType;
        component._id = componentId;
        c[componentId] = component;
        return component;
    }
}
