import { simd, threads } from 'wasm-feature-detect';
import { WonderlandEngine } from './engine.js';
import { WASM } from './wasm.js';
import * as API from './wonderland.js';
export * from './wonderland.js';
export * from './engine.js';
export * from './wasm.js';
function loadScript(scriptURL) {
    return new Promise((res, rej) => {
        const s = document.createElement('script');
        const node = document.body.appendChild(s);
        s.onload = () => {
            document.body.removeChild(node);
            res();
        };
        s.onerror = (e) => {
            document.body.removeChild(node);
            rej(e);
        };
        s.src = scriptURL;
    });
}
/**
 * Finds whether simd & threading are supported or not.
 *
 * @returns An object containing boolean for simd and thread.
 */
async function detectFeatures() {
    let [simdSupported, threadsSupported] = await Promise.all([simd(), threads()]);
    if (simdSupported) {
        console.log('WASM SIMD is supported');
    }
    else {
        console.warn('WASM SIMD is not supported');
    }
    if (threadsSupported) {
        if (self.crossOriginIsolated) {
            console.log('WASM Threads is supported');
        }
        else {
            console.warn('WASM Threads is supported, but the page is not crossOriginIsolated, therefore thread support is disabled.');
        }
    }
    else {
        console.warn('WASM Threads is not supported');
    }
    threadsSupported = threadsSupported && self.crossOriginIsolated;
    return {
        simdSupported,
        threadsSupported,
    };
}
/**
 * Load the runtime using the WASM and JS files.
 *
 * @param runtime The runtime base string, e.g,: `WonderlandRuntime-loader-physx`.
 * @param options Options to modify the loading behaviour.
 *
 * @returns A promise that resolves when the engine is ready to be used.
 */
export async function loadRuntime(runtime, options = {}) {
    const { simdSupported, threadsSupported } = await detectFeatures();
    const { simd = simdSupported, threads = threadsSupported, physx = false, loader = false, } = options;
    const variant = [];
    if (loader)
        variant.push('loader');
    if (physx)
        variant.push('physx');
    if (simd)
        variant.push('simd');
    if (threads)
        variant.push('threads');
    const variantStr = variant.join('-');
    let filename = runtime;
    if (variantStr)
        filename = `${filename}-${variantStr}`;
    const r = await fetch(filename + '.wasm');
    if (!r.ok) {
        return Promise.reject('Failed to fetch runtime .wasm file');
    }
    const wasm = new WASM(threads);
    wasm.worker = `${filename}.worker.js`;
    wasm.wasm = await r.arrayBuffer();
    const engine = new WonderlandEngine(wasm);
    if (!window._WL) {
        window._WL = { runtimes: {} };
    }
    const runtimes = window._WL.runtimes;
    /* Global identifier of this runtime in `window`. */
    const runtimeGlobalId = variantStr ? variantStr : 'default';
    /* Only load the runtime if not previously loaded in the page */
    if (!runtimes[runtimeGlobalId]) {
        await loadScript(`${filename}.js`);
        runtimes[runtimeGlobalId] = window.instantiateWonderlandRuntime;
        window.instantiateWonderlandRuntime = undefined;
    }
    await runtimes[runtimeGlobalId](wasm);
    engine._init();
    engine.start();
    /* Backward compatibility. @todo Remove at 1.0.0 */
    Object.assign(engine, API);
    /* @ts-ignore */
    window.WL = engine;
    return engine;
}
